#include <pic16f628a.h>
#include <stdint.h> // For uint8_t and fellows

//#define HAS_TINY_MULTIPLY

/*
    This program is for PIC16f628a.
    It has room for:
        2048 program 14-bit bytes (All code & constant data.)
                                      from 0000 to 07FF
        224  SRAM     8-bit bytes : 80 in  20-6F      bank 0
                                    16 in  70-7F or F0-FF or 170-17F or 1F0-1FF
                                    80 in  A0-EF      bank 1
                                    48 in 120-14F     bank 2
        128  eeprom  8-bit bytes
                         Accessed only through EECON1,EECON2,EEADR,RD,WR,
                                               WREN,EEDATA  ports.
        16   I/O     bits
                Pin    Name                 Functions
                  17    RA0/AN0               DIGITAL IN&OUT / ANALOG IN
                  18    RA1/AN1               DIGITAL IN&OUT / ANALOG IN
                   1    RA2/AN2/Vref          DIGITAL IN&OUT / ANALOG IN
                   2    RA3/AN3/CMP1          DIGITAL IN&OUT / ANALOG IN
                   3    RA4/T0CKI/CMP2        DIGITAL IN&OUT (open drain output)
                                                             / TIMER0 CLOCK IN / CMP 2 OUT
                   4    RA5/^MCLR/VPP         DIGITAL IN     / PROGRAMMING MODE ENABLE
                  15    RA6/OSC2/CLKOUT       DIGITAL IN&OUT / OSC
                  16    RA7/OSC1/CLKIN        DIGITAL IN&OUT / OSC / CLOCK IN

                   6    RB0/INT               DIGITAL IN&OUT / INTERRUPT
                   7    RB1/RX/DT             DIGITAL IN&OUT / USART
                   8    RB2/TX/CK             DIGITAL IN&OUT / USART
                   9    RB3/CCP1              DIGITAL IN&OUT / PWM OUTPUT
                  10    RB4/PGM               DIGITAL IN&OUT
                  11    RB5                   DIGITAL IN&OUT
                  12    RB6/T1OSO/T1CKI/PGC   DIGITAL IN&OUT / TIMER1 CLOCK IN
                  13    RB7/T1OSI/PGD         DIGITAL IN&OUT / TIMER1 OSC IN


        Pins required for oscillation:
      16  RA7 = OSC1 (used for RC, or for crystal)
      15  RA6 = OSC2 (used for crystal)
        Other pins used:
      17  RA0 = LED display clock
      18  RA1 = LED display data
       1  RA2 = LED display strobe0
       9  RB3 = speaker

    Oscillator option:

    External 20 MHz oscillator
        needs two capactors like...
            crystal goes between RA6 and RA7
            both legs are also connected to GND through
            separate 15-30 pF capacitors.
    Or,
        could use RC circuit
            fr = 1 / (2*pi*R*C*sqrt(2*N))
            where N=1, fr=20MHz

            R = 56k, C =      1e-13 (0.1    pF)
            R = 220, C = 2.5577e-11 (25.577 pF)
            R =  56, C =      1e-10 (100 pF = 0.1 nF) --> 20096346 Hz

    7-segment pinout:

          INPUT
        VCC  GND
        CLK  DIO
      | STB0 stb1
        stb2 stb3
        stb4 stb5

    */



#include "songdata.h"

#define MAX_LOOPS    3
#define NUM_CHANNELS 4

static uint16_t cp;  // CurrentPointer[CurChannel]
static uint8_t  vcl; // VolumeCurveLength[CurChannel]
static uint8_t  vpp; // VolumeCurvePosition[CurChannel];

static uint8_t  CurChannel;                        // 2-bit (0-3)
static uint8_t  NoteBaseOffset[NUM_CHANNELS];      // 6-bit (24-55)
static uint8_t  VolumeAndEnv[NUM_CHANNELS];

static uint8_t  VolumeCurveLength[NUM_CHANNELS];   // sign-bit + 7-bit integer
static uint8_t  VolumeCurvePosition[NUM_CHANNELS]; // 7-bit
static uint8_t  VolumeCurveStrength[NUM_CHANNELS]; // 0..15 or -1..-15

static uint8_t  ActiveVolume[NUM_CHANNELS];        // 0..15
static uint8_t  DutyCycle[NUM_CHANNELS];           // 2-bit
static uint16_t LengthCounter[NUM_CHANNELS];       // 16-bit, because multiplied by Speed
static uint8_t  CurrentNote[NUM_CHANNELS];         // 7-bit (24-70)
static uint8_t  KeyOnDelay[NUM_CHANNELS];
static uint8_t  SongSpeed[NUM_CHANNELS];

static uint16_t CurrentPhase[NUM_CHANNELS] = {0,0,0,0};
static uint16_t CurrentRate[NUM_CHANNELS]  = {0,0,0,0};

// Song is ~700 lines long. These pointers need only be 10-bit.
static uint16_t CurrentPointer[NUM_CHANNELS] = {PTR0,PTR1,PTR2,PTR3};
static uint8_t  LoopCounters[MAX_LOOPS][NUM_CHANNELS]; // 4-bit
static uint16_t LoopPointers[MAX_LOOPS][NUM_CHANNELS] =
    { {PTR0,PTR1,PTR2,PTR3},
      {PTR0,PTR1,PTR2,PTR3} };

// Set the __CONFIG word:
# pragma config WDTE=OFF,FOSC=HS,LVP=ON,CPD=OFF,CP=OFF,PWRTE=ON,MCLRE=OFF

#define U (uint8_t)
#define W (uint16_t)

static uint8_t GetByte() __attribute__((noinline));
static uint8_t GetByte()
{
    return songdata[ W(cp++) ];
}


static uint8_t CalculateVolume(uint8_t v/*channel volume*/, uint8_t wave/*momentary wave value*/)
{
    // Return value: approximately (wave-7.5) * v/2  (-52..+52)
#ifdef HAS_TINY_MULTIPLY
    return ((v & 15) * (signed char)((wave<<4) - 0x78)) / 34;
#else
    uint8_t result = 0;

    if(!(wave & 8)) wave ^= U 7;

    if(U(v & 1)) result += U(wave&0x07); // 0x08; 0..7
    wave <<= 1;

    if(U(v & 2)) result += U(wave&0x0E); // 0x10; 0..14 or 0..21
    wave <<= 1;

    if(U(v & 4)) result += U(wave&0x1C); // 0x20; 0..28 or 0..49
    wave <<= 1;

    if(U(v & 8)) result += U(wave&0x38); // 0x40; 0..56 or 0..105
    result >>= 1;                    // 0..52

    if(U(wave & 0x40)) return result;
    return -result;
#endif
}

static const uint8_t RndData[64] =
{
    0xC8,0xE8,0xE7,0xCB,0x31,0x99,0xB0,0x10,0x3F,0xED,0xC0,0x19,0x76,0xF3,0x5B,0x45,
    0x13,0xE7,0x9E,0x8C,0xDF,0x30,0xC1,0x3C,0x72,0x5F,0xC1,0x83,0x3A,0x11,0x95,0x4E,
    0x8D,0x9C,0x8F,0xE3,0x4F,0x18,0x3B,0x3B,0x32,0x3F,0xF1,0xDF,0xFF,0x69,0x40,0xFA,
    0x18,0xE8,0x64,0x26,0x8E,0xB8,0xEE,0xCD,0xD1,0x81,0x2E,0xCB,0x9B,0x63,0x9B,0x2E,/*
    0xBB,0xDF,0x29,0x5B,0x2E,0x38,0x50,0x2C,0x87,0xCA,0xD9,0x58,0x4E,0x6F,0xDA,0x01,
    0x78,0x05,0x89,0x16,0xC8,0x77,0x60,0x28,0x07,0xFB,0x6F,0x9A,0xB1,0xDA,0x40,0x4F,
    0xD4,0x82,0x63,0x29,0x05,0xFD,0x3C,0x40,0x9B,0xF6,0x6C,0x2E,0x3A,0x44,0x24,0xCF,
    0xF0,0xC5,0xFC,0x69,0x35,0xE5,0x38,0x27,0xEB,0x91,0xBC,0x08,0x62,0x31,0xAA,0xD0*/
};

static const uint8_t LEDfont[16] =
{
    //           76543210
    0x3F ,//   0b00111111, // 0  
    0x6  ,//   0b00000110, // 1
    0x5B ,//   0b01011011, // 2
    0x4F ,//   0b01001111, // 3
    0x66 ,//   0b01100110, // 4
    0x6D ,//   0b01101101, // 5
    0x7D ,//   0b01111101, // 6
    0x27 ,//   0b00100111, // 7
    0x7F ,//   0b01111111, // 8
    0x6F ,//   0b01101111, // 9
    0x77 ,//   0b01110111, // A
    0x7C ,//   0b01111100, // B
    0x39 ,//   0b00111001, // C
    0x5E ,//   0b01011110, // D
    0x79 ,//   0b01111001, // E
    0x71  //   0b01110001  // F
};

enum { LED_PinClock = 0x01, LED_PinData = 0x02, LED_PinStrobe = 0x04,
       LED_ColorRed = 1, LED_ColorGreen = 2, LED_ColorBoth = 3 };

static void LED_SendByte(uint8_t byte)
{
    for(uint8_t c=8; ; byte>>=1)
    {
        PORTA &= ~LED_PinClock;
        if(byte & 1) PORTA |= LED_PinData;
        else         PORTA &= ~LED_PinData;
        PORTA |=  LED_PinClock;
        if(!--c) break;
    }
}
static void LED_SendCommand(uint8_t command)
{
    PORTA &= ~LED_PinStrobe;
    LED_SendByte(command);
    PORTA |=  LED_PinStrobe;
}

static void LED_SendData(uint8_t index, uint8_t data)
{
    LED_SendCommand(0x44);
    PORTA &= ~LED_PinStrobe;
    LED_SendByte(0xC0 | index);
    LED_SendByte(data);
    PORTA |= LED_PinStrobe;
}

static void LED_SetChar(uint8_t index, uint8_t ch) // ch=0..15. If >=16, show blank.
{
    LED_SendData( index, U(ch&0x10) ? U 0x00 : U LEDfont[ U(ch&0x0F) ] );
}

static void LED_Init(void)
{
    PORTA |= LED_PinStrobe;
    PORTA |= LED_PinClock;
    LED_SendCommand(0x40);
    LED_SendCommand(0x8D); // 0x08 = activate display, 0x05 = intensity 5
    PORTA &= ~LED_PinStrobe;
    LED_SendByte(0xC0);
    for(uint8_t c=16; --c; )  LED_SendByte(0x00);
    PORTA |= LED_PinStrobe;
}

static uint16_t amplitude = 0;
static void UpdateLEDdisplay(void) __attribute__((noinline));
static void UpdateLEDdisplay(void)
{
    static uint16_t tick_counter = 0;
    LED_SetChar( 8, U(tick_counter >> 12) & 0x0F );
    LED_SetChar(10, U(tick_counter >>  8) & 0x0F );
    LED_SetChar(12, U(tick_counter >>  4) & 0x0F );
    LED_SetChar(14, U(tick_counter >>  0) & 0x0F );

    // Decide upon color and amplitude.
    // Color = LED_ColorGreen for scaled amplitudes 0..4
    // Color = LED_ColorBoth  for scaled amplitudes 5..6
    // Color = LED_ColorRed   for scaled amplitudes 7..7
    // Because we are playing at about 13 kHz rate and the
    // UpdateLEDdisplay function is called 76 times in a second,
    // this means our amplitude is maximally about 22000.
    // To scale that into 0..7 range, we need to multiply
    // it by 1.5 and divide by 2^12. If result>7, make 7.
    uint8_t amp   = (W(amplitude >> 1) + amplitude) >> 4;
    uint8_t color = LED_ColorGreen;
    if(amp >= 5) color = (amp >= 7 ? U LED_ColorRed : U LED_ColorBoth);

    for(uint8_t c=0; c<8; ++c)
        LED_SendData( U(c<<1), amp>=c ? color : U 0);

    amplitude = 0;
    ++tick_counter;
}

int main(void)
{
    // This constant controls the song playback rate.
    // It is calculated with round(720 * 16 / 76.2939),
    // where 720 Hz is the desired playback tick rate
    // and 76.2939 Hz is the interval between TMR0
    // wrapping back to 0 (which comes from 20e6/4/256/256).
    enum { TIMING_NOMINATOR = 151 };

    // This constant controls the pitch of the melody.
    // It is calculated with 121.6796875 / 19531.25
    // where 19531.25 Hz is the interval between TMR0 increments
    // (which on the PIC16F628A comes from 20e6/4/256),
    // and 121.6796875 happens to be some finetuning.
    #define FREQ_SCALE 0.00623

    // Amplitudes. Chosen carefully in such manner that the sample never clips.
    enum { SQUARE0 = 3, SQUARE1 = 13, TRIVOLUME = 12, NOISE0 = 8, NOISE1 = 13, OFFSET = 0x66 };

#if !defined(COMPILING_REAL_PIC)

    #define SEND_SAMPLE(n) \
        putchar(n)

#else

    /** BEGIN PIC16F628A CONFIGURATION **/

    TRISA = 0x00; // PORTA: Pins 6 & 7 for OSC (in), 0-2 for LED display (out)
    TRISB = 0x00; // PORTB is all outputs (and pin 3 may be PWM)

    // Set up presaler to 1:64, resulting in 256*64=16384 cycles between interrupts.
    // At 4 MHz master clock, this results in 16.384 ms interval between interrupts,
    // which corresponds rougly to 60 Hz.
    // At 20 MHz master clock, we could use 1:256, which gets about 76 Hz rate.
    //                         TMR0 resolution is about 19 kHz. Sufficient for us.

    OPTION_REG = 0x07; // Clear T0CS, Set prescaler to 1:256 (if 7). Or 1:64 (if 5)
    INTCON     = 0x00; // No interrupts.
    CMCON      = 0x07; // Disable comparators. Required on pic16F628A.
    PCON       = 0x08; // Set OSCF to high-speed mode.

    // Set up PWM:
    PR2   = 0x3F; // 78125 Hz at 8-bit resolution  (20e6/4/64)
    T2CON = 0x04; // Timer2 on, prescale 1:1
//    CCP1CON = 0x0C; // Unnecessary; repeatedly done always when sample is updated

    /** END PIC16F628A CONFIGURATION **/

    // Hardware interface: Send 8-bit PCM sample for PWM modulator.
    #define SEND_SAMPLE(n) \
        CCP1CON = ((n & 3) << 4) | 0x0C; \
        CCPR1L  = (n >> 2)
#endif

    LED_Init();

    // Loop forever.
    for(;;)
    {
        // Update the waveform on each channel.
        // Because of timer0 interrupts (and the code executed therein
        // taking different amounts of time at different times), the
        // PCM generation may occur at uneven intervals. To compensate,
        // we take a reading from timer0, to see how many cycles we are
        // behind from what we are supposed to do. TMR0 will change
        // 20e6/4/256 = 19531.25 times per second, which is more than
        // enough for our purposes (simulation tests indicate we are
        // getting about 12500 Hz playback).
        static uint8_t  prevmul = 0, flag, speed;

        enum { FLAG_TICK = 0x01, FLAG_CURVECHANGED = 0x02 };

#if !defined(COMPILING_REAL_PIC)
        ++TMR0; // Manually increase TMR0 because there's
                // no background device to do so
#endif

        if(prevmul == TMR0) continue;

        // Catch up as many cycles as we missed
        do {
            if(!prevmul) flag |= FLAG_TICK; // Did we hit the 76.2939 Hz interval?

            // Time to execute new song row?
            CurrentPhase[0] += CurrentRate[0];
            CurrentPhase[1] += CurrentRate[1];
            CurrentPhase[2] += CurrentRate[2];
            CurrentPhase[3] += CurrentRate[3];
        } while(U ++prevmul != TMR0);

        /* This function returns a waveform for a given channel
         * and its phase: Square, triangle or noise.
         * For channels 0 & 1:
         *        Square. Uses: CurrentPhase high 3 bits, and DutyCycle.
         *                Returns: 14 or 4.
         * For channel 2:
         *        Triangle. Uses: CurrentPhase high 5 bits.
         *                  Returns 0..15.
         * For channel 3:
         *        Periodic noise. Uses: CurrentPhase, and RndData[].
         *                        Returs 0 or 7.
         */

        // One triangle channel.
        uint8_t
        sample = CalculateVolume( TRIVOLUME,
                                  U(U(CurrentPhase[2] >> 11) & 15) ^ U((CurrentPhase[2] & (16<<11)) ? U 15 : U 0) );

        // One noise channel.
        sample += CalculateVolume( ActiveVolume[3],
                                   U(U((U RndData[  U(U(CurrentPhase[3] >> 2) & 0x3F) ]) >> U(CurrentPhase[3] >> 13)) & 1)
                                   ? U NOISE1 : U NOISE0);

        // Two square channels.
        sample += CalculateVolume( ActiveVolume[1],
                                   U(CurrentPhase[1] >> U(12 + DutyCycle[1])) ? U SQUARE1 : U SQUARE0 );

        sample += CalculateVolume( ActiveVolume[0],
                                   U(CurrentPhase[0] >> U(12 + DutyCycle[0])) ? U SQUARE1 : U SQUARE0 );

        // Normalize to 0..255 range (no negative numbers)
        sample += OFFSET;

        // Update approximate amplitude (statistics only for user interface)
        amplitude += U (sample & 0x7F);

        // Signal the value.
        SEND_SAMPLE(sample);

        if(flag & FLAG_TICK)
        {
            speed += TIMING_NOMINATOR;

            UpdateLEDdisplay();

            // Update the song state
            for(CurChannel=0; CurChannel!=NUM_CHANNELS; ++CurChannel)
            {
                flag &= ~(FLAG_TICK | FLAG_CURVECHANGED);

                // Cache important variables
                cp  = CurrentPointer[CurChannel];
                vcl = VolumeCurveLength[CurChannel];
                vpp = VolumeCurvePosition[CurChannel];

                uint16_t new_length_counter = LengthCounter[CurChannel];

                // The length intervals used by our code are for 720 Hz. (60*12).
                // Because we are running at 76.2939 Hz, we need a factor
                // of 9.437184 for preciseness.
                // We use 151/16 = 9.4375, which is pretty close.
                // It gets 720.02411 for song speed.
                // 147456/15625 would be precise, but is inconvenient.
                int8_t frames_remain = -(speed >> 4);

                uint8_t note = CurrentNote[CurChannel];
                uint8_t kod  = KeyOnDelay[CurChannel] + frames_remain;

                // Update timing
                if(W(new_length_counter & 0x8000)) // negative?
                {
                    // Take back the part we did not consume yet.
                    frames_remain += new_length_counter;
                    new_length_counter = 0;
                }
                if(!new_length_counter)
                {
                    // Read any following commands.
                    // Stop when the first note/pause is encountered.
                    uint8_t length_counter = 6;
                    uint8_t note_delay     = 0;
                    for(;;)
                    {
                        uint8_t byte  = GetByte();
                        if( U(byte & 0x80))
                        {
                            if( U(byte < U 0xE0))
                            {
                                length_counter = U(U length_counter << U(U(byte >> 4) & U 7));
                                // Length can be: 6,12,24,48,96,192
                                //            Or: 9,18,36,72,144,288 (288 is too large though)
                                //            Or: 4,8,16,32,64,128
                            NoteCommon:;
                              #ifdef HAS_TINY_MULTIPLY
                                new_length_counter = length_counter * SongSpeed[CurChannel];
                                kod                = note_delay     * SongSpeed[CurChannel] - 1;
                              #else
                                uint8_t s = SongSpeed[CurChannel];
                                uint16_t length_counter_16bit = length_counter;
                                for(kod=-1 ; ; )
                                {
                                    if(U(s & 1))
                                    {
                                        new_length_counter += length_counter_16bit;
                                        kod                += note_delay;
                                    }
                                    if(!U(s >>= 1)) break;
                                    length_counter_16bit <<= 1;
                                    note_delay           <<= 1;
                                }
                              #endif
                                note = byte & 0x0F;
                                if(note) note += NoteBaseOffset[CurChannel];
                                break;
                            }
                            // byte >= 0xE0, i.e. SetNoteBaseOffset

                            uint8_t tmp = U(byte & 0x1F) + 24 - 1;
                            if(byte == 0xFF) tmp = GetByte()-1;
                            NoteBaseOffset[CurChannel] = tmp;
                        }
                        else
                        {
                            uint8_t command = byte & 7;
                            byte >>= 3; // 4-bit parameter
                            if( U(command & 4))
                            {
                                // 4..7
                                if( U(command & 2))
                                {
                                    // 6..7
                                    if( U(command & 1))  // 7: Set Volume Curve (next 2 bytes)
                                    {
                                        vcl = GetByte();
                                        if( U(vcl & 0x80) ) byte = -byte;
                                        VolumeCurveStrength[CurChannel] = byte;
                                        flag |= FLAG_CURVECHANGED;
                                    }
                                    else // 6: Set Volume And Env (next byte)
                                    {
                                        if(byte)
                                            SongSpeed[CurChannel] = byte; // Set Speed
                                        else
                                            VolumeAndEnv[CurChannel] = GetByte();
                                    }
                                }
                                else
                                {
                                    // 4..5
                                    uint8_t loop_id = 1; // subloop
                                #if MAX_LOOPS >= 3
                                    if( U(command & 1) ) // 5: End Subloop2
                                    {
                                        ++loop_id;
                                        //passthru
                                    }
                                #endif
                                    // 4: End loop
                                #if MAX_LOOPS >= 2
                                    if(!byte)
                                        --loop_id;
                                    else // 0=unconditional mainloop
                                    {
                                        uint8_t lf = LoopCounters[loop_id][CurChannel];
                                        if( U(lf++) == byte) lf = 0; // done
                                        LoopCounters[loop_id][CurChannel] = lf;
                                        if(!lf) goto didloop;
                                    }
                                #else
                                    --loop_id;
                                #endif
                                    // Goto loop
                                    cp = LoopPointers[loop_id][CurChannel];
                                 didloop:;
                                }
                            }
                            else
                            {
                                // 0..3
                                if( U(command & 2) )
                                {
                                    // 2..3
                                    if( U(command & 1) ) // 3: Use fixed 72-length note.
                                    {
                                        length_counter = 72;
                                        goto NoteCommon;
                                    }
                                    if( U(byte & 8) )
                                        if( U(byte & 4) ) // Begin loop
                                        {
                                            // It does not matter whether we make the anchor
                                            // at the beginning of this instruction or after it.
                                            LoopPointers[ U(byte & 3) ][CurChannel] = cp;
                                        }
                                        else // Set duty cycle
                                            DutyCycle[CurChannel] = U(1 + U(byte & 3));
                                    else
                                        note_delay = byte; // Delay next note by some length-ticks
                                }
                                else
                                {
                                    // 0: Interpret next note at 3/2 its nominal length (9)
                                    // 1: Triplet timing (next note at 2/3 its nominal length) (4)
                                    if( U(command&1) )
                                        length_counter = U(6*2/3);
                                    else
                                        length_counter = U(6*3/2);
                                }
                            }
                        }
                    }
                }

                // Check if a note is due to be played right now
                if( U(kod & 0x80) && !U(note & 0x80))
                {
                    // Begin note.
                    uint16_t period = 0;
                    if(note)
                    {
                        static const uint16_t RateTable[24] =
                        {
                            // These constants are 65535 * 2^(n/12)/ with
                            // -11<=n<=0, representing the equal temperament
                            // tuning with maximum 16-bit precision.
                            34716*FREQ_SCALE*0.5, 36780*FREQ_SCALE*0.5, 38967*FREQ_SCALE*0.5,
                            41284*FREQ_SCALE*0.5, 43739*FREQ_SCALE*0.5, 46340*FREQ_SCALE*0.5,
                            49096*FREQ_SCALE*0.5, 52015*FREQ_SCALE*0.5, 55108*FREQ_SCALE*0.5,
                            58385*FREQ_SCALE*0.5, 61857*FREQ_SCALE*0.5, 65535*FREQ_SCALE*0.5,
                            34716*FREQ_SCALE*1.0, 36780*FREQ_SCALE*1.0, 38967*FREQ_SCALE*1.0,
                            41284*FREQ_SCALE*1.0, 43739*FREQ_SCALE*1.0, 46340*FREQ_SCALE*1.0,
                            49096*FREQ_SCALE*1.0, 52015*FREQ_SCALE*1.0, 55108*FREQ_SCALE*1.0,
                            58385*FREQ_SCALE*1.0, 61857*FREQ_SCALE*1.0, 65535*FREQ_SCALE*1.0
                        };
                        if(CurChannel == 3) note += 6;//-= 20;
                        if(CurChannel == 2) note -= 12;
                        uint8_t o = 0;
                        for(; note >= 24; note -= 24) o += 2; // div-mod
                        // With note range 24-70, the octave range is 2..5.
                        period = W((W RateTable[note]) << o);
                    }
                    CurrentRate[CurChannel] = period;
                    note |= 0x80;
                    flag |= FLAG_CURVECHANGED;
                }

                if( U(flag & FLAG_CURVECHANGED))
                {
                    // Reset volume curve
                    if(CurChannel == 2)
                    {
                        // Linear counter
                        vpp = U(VolumeAndEnv[CurChannel] & 0x7F);
                    }
                    else if(vcl)
                    {
                        ActiveVolume[CurChannel] = U(vcl & 0x80) ? U 15 : U 0;
                        vpp = 0;
                    }
                }

                // Update timing
                LengthCounter[CurChannel] = W(new_length_counter + frames_remain);

                // Update volume curve
                if(CurChannel == 2)
                {
                    // Linear counter, if enabled
                    if(!U(VolumeAndEnv[CurChannel] & 0x80))
                    {
                        // Accumulate the linear counter twice per frame
                        if(vpp && (!U --vpp || !U --vpp))
                            CurrentRate[CurChannel] = 0;
                    }
                }
                else
                {
                    // At VolumeCurveLength intervals, VolumeCurveStrength is added
                    // to the volume, clamping between 0 and (VolumeEnv & 0xF).
                    uint8_t s = ActiveVolume[CurChannel];
                    if(U(vcl & 0x7F) == U ++vpp)
                    {
                        vpp = 0;
                        // Add S (or -S if l&0x80)
                        s += VolumeCurveStrength[CurChannel];
                        // Clamp to 0..v
                        if(U(s & 0x80)) s = 0;
                    }

                    uint8_t v = VolumeAndEnv[CurChannel] & 0x0F;
                    if(s >= v)   s = v;
                    ActiveVolume[CurChannel] = s;
                }

                // Save cached variables back
                CurrentPointer[CurChannel]      = cp;
                VolumeCurveLength[CurChannel]   = vcl;
                VolumeCurvePosition[CurChannel] = vpp;
                CurrentNote[CurChannel] = note;
                KeyOnDelay[CurChannel] = kod;
            }

            speed &= 0xF;
        }
    }
}
